"""Various types and representations of graphs."""
import numpy as np
from numpy.random import default_rng


class UndirectedDependenceGraph(object):
    r"""Adjacency matrix representation using a 2d numpy array.

    Upon initialization, this class is fairly standard implementation
    of an undirected graph. However, upon calling the :meth:`make_aux`
    method, an auxilliary data structure in the form of several new
    attributes is created, which are used by
    :meth:`medil.ecc_algorithms.find_clique_min_cover` according to
    the algorithm in :cite:`Gramm_2009`.

    Attributes
    ----------


    Notes
    -----
    The algorithms for finding the minMCM via ECC contain many
    algebraic operations, so adjacency matrix representation (via
    NumPy) is most covenient.

    The diag is used to store information when the graph is reduced,
    not to indicate self loops, so it is important that diag = 1 at
    init.

    """

    def __init__(self, adj_matrix, verbose=False):
        # doesn't behave well unless input is nparray
        self.adj_matrix = adj_matrix
        self.num_vertices = np.trace(adj_matrix)
        self.max_num_verts = len(adj_matrix)
        self.num_edges = np.triu(adj_matrix, 1).sum()
        self.verbose = verbose

    def add_edges(self, edges):
        v_1s = edges[:, 0]
        v_2s = edges[:, 1]
        self.adj_matrix[v_1s, v_2s] = 1
        self.adj_matrix[v_2s, v_1s] = 1
        self.num_edges = np.triu(self.adj_matrix, 1).sum()

    def rm_edges(self, edges):
        v_1s = edges[:, 0]
        v_2s = edges[:, 1]
        self.adj_matrix[v_1s, v_2s] = 0
        self.adj_matrix[v_2s, v_1s] = 0
        self.num_edges = np.triu(self.adj_matrix, 1).sum()

    def make_aux(self):
        # this makes the auxilliary structure described in INITIALIZATION in the paper

        # find neighbourhood for each vertex
        # each row corresponds to a unique edge
        max_num_edges = self.n_choose_2(self.max_num_verts)
        self.common_neighbors = np.zeros(
            (max_num_edges, self.max_num_verts), int
        )  # init

        # mapping of edges to unique row idx
        triu_idx = np.triu_indices(self.max_num_verts, 1)
        nghbrhd_idx = np.zeros((self.max_num_verts, self.max_num_verts), int)
        nghbrhd_idx[triu_idx] = np.arange(max_num_edges)
        # nghbrhd_idx += nghbrhd_idx.T
        self.get_idx = lambda edge: nghbrhd_idx[edge[0], edge[1]]

        # reverse mapping
        u, v = np.where(np.triu(np.ones_like(self.adj_matrix), 1))
        self.get_edge = lambda idx: (u[idx], v[idx])

        # compute actual neighborhood for each edge = (v_1, v_2)
        self.nbrs = lambda edge: np.logical_and(
            self.adj_matrix[edge[0]], self.adj_matrix[edge[1]]
        )

        extant_edges = np.transpose(np.triu(self.adj_matrix, 1).nonzero())
        self.extant_edges_idx = np.fromiter(
            {self.get_idx(edge) for edge in extant_edges}, dtype=int
        )
        extant_nbrs = np.array([self.nbrs(edge) for edge in extant_edges], int)
        extant_nbrs_idx = np.array([self.get_idx(edge) for edge in extant_edges], int)

        # from paper: set of N_{u, v} for all edges (u, v)
        self.common_neighbors[extant_nbrs_idx] = extant_nbrs

        # number of cliques for each node? assignments? if we set diag=0
        # num_cliques = common_neighbors.sum(0)

        # sum() of submatrix of graph containing exactly the rows/columns
        # corresponding to the nodes in common_neighbors(edge) using
        # logical indexing:

        # make mask to identify subgraph (closed common neighborhood of
        # nodes u, v in edge u,v)
        mask = lambda edge_idx: np.array(self.common_neighbors[edge_idx], dtype=bool)

        # make subgraph-adjacency matrix, and then subtract diag and
        # divide by two to get num edges in subgraph---same as sum() of
        # triu(subgraph-adjacency matrix) but probably a bit faster
        nbrhood = lambda edge_idx: self.adj_matrix[mask(edge_idx)][:, mask(edge_idx)]
        max_num_edges_in_nbrhood = (
            lambda edge_idx: (nbrhood(edge_idx).sum() - mask(edge_idx).sum()) // 2
        )

        # from paper: set of c_{u, v} for all edges (u, v)
        self.nbrhood_edge_counts = np.array(
            [
                max_num_edges_in_nbrhood(edge_idx)
                for edge_idx in np.arange(max_num_edges)
            ],
            int,
        )

        # important structs are:
        # self.common_neighbors
        # self.nbrhood_edge_counts
        # # and fun is
        # self.nbrs

    @staticmethod
    def n_choose_2(n):
        return n * (n - 1) // 2

    def reducible_copy(self):
        return ReducibleUndDepGraph(self)

    def convert_to_nde(self, name="temp"):
        with open(name + ".nde", "w") as f:
            f.write(str(self.max_num_verts) + "\n")
            for idx, node in enumerate(self.adj_matrix):
                f.write(str(idx) + " " + str(node.sum()) + "\n")
            for v1, v2 in np.argwhere(np.triu(self.adj_matrix)):
                f.write(str(v1) + " " + str(v2) + "\n")


class ReducibleUndDepGraph(UndirectedDependenceGraph):
    def __init__(self, udg):
        self.unreduced = udg.unreduced if hasattr(udg, "unreduced") else udg
        self.adj_matrix = udg.adj_matrix.copy()
        self.num_vertices = udg.num_vertices
        self.num_edges = udg.num_edges

        self.the_cover = None
        self.verbose = udg.verbose

        # from auxilliary structure if needed
        if not hasattr(udg, "get_idx"):
            udg.make_aux()
        self.get_idx = udg.get_idx

        # need to also update these all when self.cover_edges() is called? already done in rule_1
        self.common_neighbors = udg.common_neighbors.copy()
        self.nbrhood_edge_counts = udg.nbrhood_edge_counts.copy()

        # update when cover_edges() is called, actually maybe just extant_edges?
        self.extant_edges_idx = udg.extant_edges_idx.copy()
        self.nbrs = udg.nbrs
        self.get_edge = udg.get_edge

        if hasattr(udg, "reduced_away"):  # then rule_3 wasn't applied
            self.reduced_away = udg.reduced_away

    def reset(self):
        self.__init__(self.unreduced)

    def reduzieren(self, k_num_cliques):
        # reduce by first applying rule 1 and then repeatedly applying
        # rule 2 or rule 3 (both of which followed by rule 1 again)
        # until they don't apply
        if self.verbose:
            print("\t\treducing:")
        self.k_num_cliques = k_num_cliques
        self.reducing = True
        while self.reducing:
            self.reducing = False
            self.rule_1()
            self.rule_2()
            if self.k_num_cliques < 0:
                return
            if self.reducing:
                continue
            self.rule_3()

    def rule_1(self):
        # rule_1: Remove isolated vertices and vertices that are only
        # adjacent to covered edges

        isolated_verts = np.where(self.adj_matrix.sum(0) + self.adj_matrix.sum(1) == 2)[
            0
        ]
        if len(isolated_verts) > 0:  # then Rule 1 is applied
            if self.verbose:
                print("\t\t\tapplying Rule 1...")

            # update auxilliary attributes; LEMMA 2
            self.adj_matrix[isolated_verts, isolated_verts] = 0
            self.num_vertices -= len(isolated_verts)

            # decrease nbrhood edge counts
            for vert in isolated_verts:
                open_nbrhood = self.unreduced.adj_matrix[vert].copy()
                open_nbrhood[vert] = 0
                idx_nbrhoods_to_update = np.where(self.common_neighbors[:, vert] == 1)[
                    0
                ]
                tiled = np.tile(
                    open_nbrhood, (len(idx_nbrhoods_to_update), 1)
                )  # instead of another loop
                to_subtract = np.logical_and(
                    tiled, self.common_neighbors[idx_nbrhoods_to_update]
                ).sum(1)
                self.nbrhood_edge_counts[idx_nbrhoods_to_update] -= to_subtract
                # my own addition:
                # self.nbrhood[:, vert] = 0

            # remove isolated_verts from common neighborhoods
            self.common_neighbors[:, isolated_verts] = 0

        # max_num_edges = self.n_choose_2(self.unreduced.max_num_verts)
        # mask = lambda edge_idx: np.array(self.common_neighbors[edge_idx], dtype=bool)

        # # make subgraph-adjacency matrix, and then subtract diag and
        # # divide by two to get num edges in subgraph---same as sum() of
        # # triu(subgraph-adjacency matrix) but probably a bit faster
        # nbrhood = lambda edge_idx: self.adj_matrix[mask(edge_idx)][:, mask(edge_idx)]
        # max_num_edges_in_nbrhood = lambda edge_idx: (nbrhood(edge_idx).sum() - mask(edge_idx).sum()) // 2

        # # from paper: set of c_{u, v} for all edges (u, v)
        # self.nbrhood_edge_counts = np.array([max_num_edges_in_nbrhood(edge_idx) for edge_idx in np.arange(max_num_edges)], int)
        # # assert (nbrhood_edge_counts==self.nbrhood_edge_counts).all()
        # # print(nbrhood_edge_counts, self.nbrhood_edge_counts)
        # # need to fix!!!!!!!! update isn't working; so just recomputing for now
        # # # # # # # # but actually update produces correct result though recomputing doesn't?

    def rule_2(self):
        # rule_2: If an uncovered edge {u,v} is contained in exactly
        # one maximal clique C, i.e., the common neighbors of u and v
        # induce a clique, then add C to the solution, mark its edges
        # as covered, and decrease k by one

        score = self.n_choose_2(self.common_neighbors.sum(1)) - self.nbrhood_edge_counts
        # score of n implies edge is in exactly n+1 maximal cliques,
        # so we want edges with score 0

        clique_idxs = np.where(score[self.extant_edges_idx] == 0)[0]

        if clique_idxs.size > 0:
            clique_idx = clique_idxs[-1]
            if self.verbose:
                print("\t\t\tapplying Rule 2...")
            clique = self.common_neighbors[self.extant_edges_idx[clique_idx]].copy()
            self.the_cover = (
                clique.reshape(1, -1)
                if self.the_cover is None
                else np.vstack((self.the_cover, clique))
            )
            self.cover_edges()
            self.k_num_cliques -= 1
            self.reducing = True
        # start the loop over so Rule 1 can 'clean up'
        # self.common_neighbors[clique_idxs[0]] = 0  # zero out row, to update struct? not in paper?

    def rule_3(self):
        # rule_3: Consider a vertex v that has at least one
        # prisoner. If each prisoner is connected to at least one
        # vertex other than v via an uncovered edge (automatically
        # given if instance is reduced w.r.t. rules 1 and 2), and the
        # prisoners dominate the exit,s then delete v. To reconstruct
        # a solution for the unreduced instance, add v to every clique
        # containing a prisoner of v.

        for vert, nbrhood in enumerate(self.adj_matrix):
            if nbrhood[vert] == 0:  # then nbrhood is empty
                continue
            exits = np.zeros(self.unreduced.max_num_verts, bool)
            nbrs = np.flatnonzero(nbrhood)
            for nbr in nbrs:
                if (nbrhood.astype(int) - self.adj_matrix[nbr].astype(int) == -1).any():
                    exits[nbr] = True
            # exits[j] == True iff j is an exit for vert
            # prisoners[j] == True iff j is a prisoner of vert
            prisoners = np.logical_and(~exits, self.adj_matrix[vert])
            prisoners[vert] = False  # a vert isn't its own prisoner

            # check if each exit is adjacent to at least one prisoner
            prisoners_dominate_exits = self.adj_matrix[prisoners][:, exits].sum(0)
            if prisoners_dominate_exits.all():  # apply the rule
                self.reducing = True
                if self.verbose:
                    print("\t\t\tapplying Rule 3...")

                # mark edges covered so rule_1 will delete vertex
                edges = np.array(
                    [
                        [vert, u]
                        for u in np.flatnonzero(self.adj_matrix[vert])
                        if vert != u
                    ]
                )
                edges.sort()  # so they're in triu, so that self.get_idx works
                self.cover_edges(edges)

                # keep track of deleted nodes and their prisoners for reconstructing solution
                if not hasattr(self, "reduced_away"):
                    self.reduced_away = np.zeros_like(self.adj_matrix, bool)
                self.reduced_away[vert] = prisoners
                break

    def choose_nbrhood(self):
        # only compute for existing edges
        common_neighbors = self.common_neighbors[self.extant_edges_idx]
        nbrhood_edge_counts = self.nbrhood_edge_counts[self.extant_edges_idx]
        score = self.n_choose_2(common_neighbors.sum(1)) - nbrhood_edge_counts

        # idx in reduced idx list, not from full edge list
        chosen_edge_idx = score.argmin()

        mask = common_neighbors[chosen_edge_idx].astype(bool)

        chosen_nbrhood = self.adj_matrix.copy()
        chosen_nbrhood[~mask, :] = 0
        chosen_nbrhood[:, ~mask] = 0
        if chosen_nbrhood.sum() <= mask.sum():
            raise ValueError("This nbrhood has no edges")
        return chosen_nbrhood

    def cover_edges(self, edges=None):
        if edges is None:
            # always call after updating the cover; only on single, recently added clique
            if self.the_cover is None:
                return  # self.adj_matrix

            # change edges to 0 if they're covered
            clique = self.the_cover[-1]
            covered = np.where(clique)[0]

            # trick for getting combinations from idx
            comb_idx = np.triu_indices(len(covered), 1)

            # actual pairwise combinations; ie all edges (v_i, v_j) covered by the clique
            covered_edges = np.empty((len(comb_idx[0]), 2), int)
            covered_edges[:, 0] = covered[comb_idx[0]]
            covered_edges[:, 1] = covered[comb_idx[1]]

        else:
            covered_edges = edges

        # cover (remove from reduced_graph) edges
        self.rm_edges(covered_edges)
        # update extant_edges_idx
        rmed_edges_idx = [self.get_idx(edge) for edge in covered_edges]
        extant_rmed_edges_idx = [
            edge for edge in rmed_edges_idx if edge in self.extant_edges_idx
        ]
        idx_idx = np.array(
            [np.where(self.extant_edges_idx == idx) for idx in extant_rmed_edges_idx],
            int,
        ).flatten()

        self.extant_edges_idx = np.delete(self.extant_edges_idx, idx_idx)
        # now here do all the updates to nbrs?----actually probably don't want this? see 2clique house example

        # update self.common_neighbors
        # self.common_neighbors[rmed_edges_idx] = 0   # zero out rows covered edges; maybe not necessary, since common_neighbors is (probably?) only called with extant_edges_idx?

        if self.verbose:
            print("\t\t\t{} uncovered edges remaining".format(self.num_edges))

        # if edges is None:
        #     cover_orig = self.the_cover
        #     while self.the_cover.shape[0] > 1:
        #         self.the_cover = self.the_cover[:-1]
        #         self.cover_edges()
        #     self.the_cover = cover_orig

    def reconstruct_cover(self, the_cover):
        if not hasattr(self, "reduced_away"):  # then rule_3 wasn't applied
            return the_cover

        # add the reduced away vert to all covering cliques containing at least one of its prisoners
        to_expand = np.flatnonzero(self.reduced_away.sum(1))
        for vert in to_expand:
            prisoners = self.reduced_away[vert]
            tiled_prisoners = np.tile(
                prisoners, (len(the_cover), 1)
            )  # instead of another loop
            cliques_to_update_mask = (
                np.logical_and(tiled_prisoners, the_cover).sum(1).astype(bool)
            )
            the_cover[cliques_to_update_mask, vert] = 1

        return the_cover


# class MCM(object):
# TODO: implement as large DAG adj matrix over L and M, or smaller bigraph for L-M connections and DAG adj matirx for L->L connections



import numpy as np
from medil.ecc_algorithms import find_clique_min_cover as find_cm
from medil.examples import examples


# Here are some integration tests.
# def test_find_cm_on_trivial_edgeless_graph():
#     # not so important at the moment
#     graph = np.array([[]], int)
#     cover = find_cm(graph, True)

#     graph = np.zeros((5, 5), int)
#     cover = find_cm(graph, True)


def test_find_cm_on_3_cycle():
    cycle_3 = np.ones((3, 3), dtype=int)
    cover = find_cm(cycle_3)
    assert cover.shape == (1, 3)
    assert ~np.any(cover - [1, 1, 1], axis=1)


def test_reduction_rule_1_on_3cycle_plus_isolated():
    graph = np.zeros((4, 4), dtype=int)
    graph[1:4, 1:4] = 1
    graph[0, 0] = 1

    cover = find_cm(graph)

    assert cover.shape == (2, 4)
    assert (cover[0] == [0, 1, 1, 1]).all()
    assert (cover[1] == [1, 0, 0, 0]).all()


def test_find_cm_on_examples():
    for example in examples:
        cover = find_cm(example.UDG)
        correct_cover = example.MCM
        assert cover.shape == correct_cover.shape
        for clique in correct_cover:
            assert (~np.any(cover - clique, axis=1)).any()



import pytest
import numpy as np

from medil.sample import mcm, biadj


def test_biadg():
    b = biadj(num_meas=5, num_latent=5)
    assert (b.sum(0) == 1).all()
    assert (b.sum(1) == 1).all()

    b = biadj(num_meas=5, density=0)
    assert (b.sum(0) == 1).all()
    assert b.sum(1).all()

    with pytest.raises(ValueError):
        biadj(num_meas=5, num_latent=1, one_pure_child=False)

    b = biadj(num_meas=10, one_pure_child=False)
    assert b.sum(0).all()
    assert b.sum(1).all()

    num_meas = 20
    max_edges = (num_meas * (num_meas - 1)) // 2
    for density in np.arange(0, 1.1, 0.1):
        biadj_mat = biadj(num_meas, density, one_pure_child=False)
        assert biadj_mat.sum(1).all()
        udg = (biadj_mat.T @ biadj_mat).astype(bool)
        density = np.triu(udg, 1).sum() / max_edges
        assert np.isclose(density, density)


def test_mcm():
    with pytest.raises(NotImplementedError):
        mcm(num_meas=2, parameterization="VAE")

    with pytest.raises(ValueError):
        mcm(num_meas=2, parameterization="test")

    m = mcm(num_meas=5)
    params = m.parameters
    assert params.biadj_weights[m.biadj].all()
    assert ~params.biadj_weights[~m.biadj].any()
    assert params.error_variances.all()
    assert params.error_means.all()



import numpy as np

# from medil.simulate import MicroModel as mm

# from medil.grues import InputData as data

# mic = mm(5, 0.5)
# samp = mic.sample(1000)
# grues = data(samp)
# grues.mcmc(prior=(2, 3))

# grues.optimal_uec
# mic.uec



import numpy as np

from medil.graph import UndirectedDependenceGraph
import medil.examples as ex


def test_make_aux_on_triangle():
    graph = UndirectedDependenceGraph(ex.triangle.UDG)
    graph.make_aux()

    # non-edges:
    assert (graph.common_neighbors[[2, 3, 4, 8, 9, 13]] == 0).all()
    # edges:
    assert (graph.common_neighbors[[0, 1]] == [1, 1, 1, 0, 0, 0]).all()
    assert (graph.common_neighbors[5] == [1, 1, 1, 0, 1, 0]).all()
    assert (graph.common_neighbors[6] == [0, 1, 0, 1, 1, 0]).all()
    assert (graph.common_neighbors[7] == [0, 1, 1, 1, 1, 0]).all()
    assert (graph.common_neighbors[10] == [0, 1, 1, 0, 1, 1]).all()
    assert (graph.common_neighbors[11] == [0, 0, 1, 0, 1, 1]).all()
    assert (graph.common_neighbors[12] == [0, 1, 0, 1, 1, 0]).all()
    assert (graph.common_neighbors[14] == [0, 0, 1, 0, 1, 1]).all()

    assert (
        graph.nbrhood_edge_counts == [3, 3, 0, 0, 0, 5, 3, 5, 0, 0, 5, 3, 3, 0, 3]
    ).all()


def test_reduce_rule_1_on_isolated_plus_3cycle():
    graph = np.zeros((4, 4), dtype=int)  # init
    graph[1:4, 1:4] = 1  # add 3cycle

    # correct answer (assuming make_aux is correct)
    correct = UndirectedDependenceGraph(graph)
    correct.make_aux()

    # to test
    graph[0, 0] = 1  # add isolated vert
    being_tested = UndirectedDependenceGraph(graph).reducible_copy()
    being_tested.rule_1()

    assert (correct.num_vertices == being_tested.num_vertices).all()
    assert (correct.common_neighbors == being_tested.common_neighbors).all()
    assert (correct.nbrhood_edge_counts == being_tested.nbrhood_edge_counts).all()


def test_reduce_rule_2_3cycle():
    init = np.ones((3, 3), dtype=int)

    graph = UndirectedDependenceGraph(init).reducible_copy()
    graph.k_num_cliques = 1
    graph.rule_2()

    assert graph.k_num_cliques == 0
    assert graph.the_cover.shape == (1, 3)
    assert ([1, 1, 1] == graph.the_cover).all()


def test_cover_edges():
    triangle_udg = ex.triangle.UDG

    clique = np.array([1, 1, 1, 0, 0, 0], int)

    being_tested = UndirectedDependenceGraph(triangle_udg).reducible_copy()
    being_tested.the_cover = clique.reshape(1, 6)
    being_tested.cover_edges()

    assert (being_tested.adj_matrix[0] == [1, 0, 0, 0, 0, 0]).all()


def test_reduce_rule_3_example_from_paper():
    graph = np.array(
        [
            [1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0],
            [1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0],
            [1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [1, 1, 1, 0, 1, 0, 1, 1, 1, 0, 0],
            [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 1],
            [0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1],
        ],
        int,
    )

    graph = UndirectedDependenceGraph(np.array(graph)).reducible_copy()

    # graph.verbose = True
    graph.rule_3()

    # correct info stored for reconstruction
    assert graph.reduced_away.sum(1)[4]

    # adj matrix updated correctly
    assert graph.adj_matrix.sum(0)[4] == 1
    assert graph.adj_matrix.sum(1)[4] == 1


def test_reconstruct_rule_3_example_from_paper():
    graph = np.array(
        [
            [1, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0],
            [1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0],
            [0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0],
            [1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [1, 1, 1, 0, 1, 0, 1, 1, 1, 0, 0],
            [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 1, 0, 1, 1, 1, 1, 1],
            [0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1],
        ],
        int,
    )

    graph = UndirectedDependenceGraph(np.array(graph)).reducible_copy()

    # graph.verbose = True
    graph.rule_3()
    graph.reduzieren(7)

    # reduced_away vert isn't in cover
    assert graph.the_cover.sum(0)[4] == 0

    cover = graph.the_cover
    # correct cover otherwise
    correct_reduced_cover = np.array(
        [
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1],
            [0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0],
            [0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
            [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            [1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
            [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]
    )

    assert (~np.any(cover - correct_reduced_cover[0], axis=1)).any()
    assert (~np.any(cover - correct_reduced_cover[1], axis=1)).any()
    assert (~np.any(cover - correct_reduced_cover[2], axis=1)).any()
    assert (~np.any(cover - correct_reduced_cover[3], axis=1)).any()
    assert (~np.any(cover - correct_reduced_cover[4], axis=1)).any()
    assert (~np.any(cover - correct_reduced_cover[5], axis=1)).any()
    assert (~np.any(cover - correct_reduced_cover[6], axis=1)).any()

    cover = graph.reconstruct_cover(graph.the_cover)
    correct_recon_cover = np.array(
        [
            [0, 0, 0, 0, 1, 0, 1, 1, 1, 0, 0],
            [0, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0],
            [1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0],
        ]
    )

    assert (~np.any(cover - correct_recon_cover[0], axis=1)).any()
    assert (~np.any(cover - correct_recon_cover[1], axis=1)).any()
    assert (~np.any(cover - correct_recon_cover[2], axis=1)).any()
    assert (~np.any(cover - correct_reduced_cover[0], axis=1)).any()
    assert (~np.any(cover - correct_reduced_cover[1], axis=1)).any()
    assert (~np.any(cover - correct_reduced_cover[4], axis=1)).any()
    assert (~np.any(cover - correct_reduced_cover[5], axis=1)).any()


# def test_reduce_rule_3_real_data():
# results = np.load("/home/alex/Projects/mcm_paper/uai_2020/data_analysis/monte_carlo_test_results_1000.npz")
# all_deps = results['deps']

# deps = all_deps[2:63, 2:63]

# c0_idx = [2, 3, 15, 17, 19, 29, 33, 39, 49, 52, 54, 55]
# c0_deps = deps[:, c0_idx][c0_idx, :]

# graph = UndirectedDependenceGraph(np.array(c0_deps, int)).reducible_copy()



import math

import torch
from torch import nn
from torch.nn.parameter import Parameter


class VariationalAutoencoder(nn.Module):
    def __init__(self, m, n, mask):
        super(VariationalAutoencoder, self).__init__()
        self.encoder = Encoder(m, n)
        self.decoder = Decoder(m, n, mask)

    def forward(self, x):
        mu, logvar = self.encoder(x)
        latent = self.latent_sample(mu, logvar)
        x_recon, logcov = self.decoder(latent)

        return x_recon, logcov, mu, logvar

    def latent_sample(self, mu, logvar):
        # the re-parameterization trick
        if self.training:
            std = logvar.mul(0.5).exp_()
            eps = torch.empty_like(std).normal_()
            return eps.mul(std).add_(mu)
        else:
            return mu


class Block(nn.Module):
    def __init__(self, m, n):
        super(Block, self).__init__()
        self.input_dim = n
        self.latent_dim = m
        self.output_dim = n


class Encoder(Block):
    def __init__(self, m, n):
        super(Encoder, self).__init__(m, n)

        # first encoder layer
        self.inter_dim = self.input_dim
        self.enc1 = nn.Linear(in_features=self.input_dim, out_features=self.inter_dim)

        # second encoder layer
        self.enc2 = nn.Linear(in_features=self.inter_dim, out_features=self.inter_dim)

        # map to mu and variance
        self.fc_mu = nn.Linear(in_features=self.inter_dim, out_features=self.latent_dim)
        self.fc_logvar = nn.Linear(
            in_features=self.inter_dim, out_features=self.latent_dim
        )

    def forward(self, x):
        # encoder layers
        inter = torch.relu(self.enc1(x))
        inter = torch.relu(self.enc2(inter))

        # calculate mu & logvar
        mu = self.fc_mu(inter)
        logvar = self.fc_logvar(inter)

        return mu, logvar


class Decoder(Block):
    def __init__(self, m, n, mask):
        super(Decoder, self).__init__(m, n)

        # decoder layer -- estimate mean
        self.dec_mean = SparseLinear(
            in_features=self.latent_dim, out_features=self.output_dim, mask=mask
        )

        # decoder layer -- estimate log-covariance
        self.fc_logcov = SparseLinear(
            in_features=self.latent_dim, out_features=self.output_dim, mask=mask
        )

    def forward(self, z):
        # linear layer
        mean = self.dec_mean(z)
        logcov = self.fc_logcov(z)

        return mean, logcov


class SparseLinear(nn.Module):
    def __init__(
        self, in_features, out_features, mask, bias=True, device=None, dtype=None
    ):
        factory_kwargs = {"device": device, "dtype": dtype}
        super(SparseLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.mask = mask
        self.weight = Parameter(
            torch.empty((out_features, in_features), **factory_kwargs)
        )

        if bias:
            self.bias = Parameter(torch.empty(out_features, **factory_kwargs))
        else:
            self.register_parameter("bias", None)
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, input):
        # masked linear layer
        return nn.functional.linear(input, self.weight * self.mask, self.bias)

    def extra_repr(self):
        return "in_features={}, out_features={}, bias={}".format(
            self.in_features, self.out_features, self.bias is not None
        )



import numpy as np

from medil.evaluate import sfd


def test_sfd():
    true_mcm = np.array([[True, True, True, False], [False, True, True, True]])
    est_mcm1 = np.array([[True, True, True, False], [False, False, True, True]])
    est_mcm2 = np.array(
        [
            [True, True, False, False],
            [True, False, True, False],
            [False, True, False, True],
            [False, False, True, True],
        ]
    )

    assert sfd(true_mcm, est_mcm1) == (5, 1)
    assert sfd(true_mcm, est_mcm2) == (10, 1)
    assert sfd(est_mcm1, est_mcm2) == (7, 2)



"""MeDIL causal model base class and a preconfigured NCFA class."""

from datetime import datetime
import os
from pathlib import Path
import pickle
import warnings

import numpy as np
from numpy.random import default_rng
import numpy.typing as npt
from scipy.linalg import sqrtm
from scipy.optimize import minimize
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler as sc
import torch
from torch.utils.data import DataLoader, TensorDataset


from .ecc_algorithms import find_heuristic_1pc
from .independence_testing import estimate_UDG
from .vae import VariationalAutoencoder


class MedilCausalModel(object):
    """Base class using principle of polymorphism to establish common
    interface for derived parametric estimators.
    """

    def __init__(
        self,
        biadj: npt.NDArray = np.array([]),
        udg: npt.NDArray = np.array([]),
        one_pure_child: bool = True,
        rng=default_rng(0),
    ) -> None:
        self.biadj = biadj
        self.udg = udg
        self.one_pure_child = one_pure_child
        self.rng = rng

    def fit(self, dataset: npt.NDArray) -> "MedilCausalModel":
        raise NotImplementedError

    def sample(self, sample_size: int) -> npt.NDArray:
        raise NotImplementedError


class Parameters(object):
    "Different parameterizations of MeDIL causal Models."

    def __init__(self, parameterization: str) -> None:
        self.parameterization = parameterization

        if parameterization == "Gaussian":
            self.error_means = np.array([])
            self.error_variances = np.array([])
            self.biadj_weights = np.array([])
        elif parameterization == "VAE":
            raise NotImplementedError

    def __str__(self) -> str:
        return "\n".join(
            f"parameters.{attr}: {val}" for attr, val in vars(self).items()
        )


class GaussianMCM(MedilCausalModel):
    """A linear MeDIL causal model with Gaussian random variables."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.parameters = Parameters("Gaussian")

    def fit(self, dataset: npt.NDArray) -> "GaussianMCM":
        """Fit a Gaussian MCM to a dataset with constraint-based
        structure learning and least squares parameter estimation."""
        self.dataset = dataset
        if self.biadj.size == 0:
            self._compute_biadj()

        self.parameters.error_means = self.dataset.mean(0)

        cov = np.cov(self.dataset, rowvar=False)

        num_weights = self.biadj.sum()
        num_err_vars = self.biadj.shape[1]

        def _objective(weights_and_err_vars):
            weights = weights_and_err_vars[:num_weights]
            err_vars = weights_and_err_vars[num_weights:]

            biadj_weights = np.zeros_like(self.biadj, float)
            biadj_weights[self.biadj] = weights

            return (
                (cov - biadj_weights.T @ biadj_weights - np.diagflat(err_vars)) ** 2
            ).sum()

        result = minimize(_objective, np.ones(num_weights + num_err_vars))
        if not result.success:
            warnings.warn(f"Optimization failed: {result.message}")

        self.parameters.error_variances = result.x[num_weights:]

        self.parameters.biadj_weights = np.zeros_like(self.biadj, float)
        self.parameters.biadj_weights[self.biadj] = result.x[:num_weights]

        return self

    def _compute_biadj(self):
        """Constraint-based structure learning."""
        if self.udg.size == 0:
            self._estimate_udg()
        self.biadj = find_heuristic_1pc(self.udg)

    def _estimate_udg(self):
        """Constraint-based structure learning."""
        samp_size = len(self.dataset)
        cov = np.cov(self.dataset, rowvar=False)
        corr = np.corrcoef(self.dataset, rowvar=False)
        inner_numerator = 1 - cov * corr  # should never be <= 0?
        inner_numerator = inner_numerator.clip(min=0.00001)
        inner_numerator[np.tril_indices_from(inner_numerator)] = 1
        udg_triu = np.log(inner_numerator) < (-np.log(samp_size) / samp_size)
        udg = udg_triu + udg_triu.T
        self.udg = udg

    def sample(self, sample_size: int) -> npt.NDArray:
        """Sample a dataset from a GaussianMCM, after structure and
        parameters have been specified or estimated."""
        num_latent = len(self.biadj)
        latent_sample = self.rng.multivariate_normal(
            np.zeros(num_latent), np.eye(num_latent), sample_size
        )
        error_sample = self.rng.multivariate_normal(
            self.parameters.error_means,
            np.diagflat(self.parameters.error_variances),
            sample_size,
        )
        sample = latent_sample @ self.parameters.biadj_weights + error_sample
        return sample


class NeuroCausalFactorAnalysis(MedilCausalModel):
    """A MeDIL causal model represented by a deep generative model."""

    def __init__(
        self,
        seed: int = 0,
        dof: int = 0,
        path: str = "trained_ncfa/",
        verbose: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        Path(path).mkdir(exist_ok=True)
        self.path = path
        self.verbose = verbose
        self.seed = seed
        self.hyperparams = {
            "heuristic": True,
            "method": "xicor",
            "alpha": 0.05,
            "dof": dof,
            "batch_size": 128,
            "num_epochs": 200,
            "lr": 0.005,
            "beta": 1,
            "num_valid": 1000,
        }
        self.parameters = Parameters("vae")
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    def log(self, entry: str) -> None:
        time_stamped_entry = f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')} {entry}"
        with open(f"{self.path}training.log", "a") as log_file:
            log_file.write(time_stamped_entry + "\n")
        if self.verbose:
            print(time_stamped_entry)

    def fit(self, dataset: npt.NDArray) -> "NeuroCausalFactorAnalysis":
        self.dataset = dataset
        self.doffed = self.assign_dof()

        standardized = sc().fit_transform(dataset)
        train_split, valid_split = train_test_split(
            standardized, train_size=0.7, random_state=self.seed
        )

        train_loader = self._data_loader(train_split)
        valid_loader = self._data_loader(valid_split)

        np.random.seed(self.seed)

        model_recon, loss_recon, error_recon = self._train_vae(
            train_loader, valid_loader
        )
        torch.save(model_recon, os.path.join(self.path, "model_recon.pt"))
        with open(os.path.join(self.path, "loss_recon.pkl"), "wb") as handle:
            pickle.dump(loss_recon, handle, protocol=pickle.HIGHEST_PROTOCOL)
        with open(os.path.join(self.path, "error_recon.pkl"), "wb") as handle:
            pickle.dump(error_recon, handle, protocol=pickle.HIGHEST_PROTOCOL)
        return self

    def assign_dof(self) -> npt.NDArray:
        """Assign degrees of freedom (latent variables) of VAE to
        latent factors from causal structure learning.
        """
        if self.biadj.size == 0:
            self._compute_biadj()

        num_cliques, num_meas = self.biadj.shape
        if self.hyperparams["dof"] == 0:
            # then default to 3x num_meas overcomplete
            self.dof = num_meas * 3
        elif self.dof < num_cliques:
            warnings.warn(
                f"Input `deg_of_freedom={self.dof}` is less than the {num_cliques} required for the estimated causal structure. `deg_of_freedom` increased to {num_cliques} to compensate."
            )
        self.dof = num_cliques

        latents_per_clique = np.ones(num_cliques, int) * (self.dof // num_cliques)

        for _ in range(2):
            remainder = self.dof - latents_per_clique.sum()
            latents_per_clique[np.argsort(latents_per_clique)[0:remainder]] += 1

        redundant_biadj_mat = np.repeat(self.biadj, latents_per_clique, axis=0)

        return redundant_biadj_mat

    def _compute_biadj(self):
        if self.udg.size == 0:
            self._estimate_udg()
        self.biadj = find_heuristic_1pc(self.udg)

    def _estimate_udg(self):
        self.udg, pvals = estimate_UDG(
            self.dataset,
            method=self.hyperparams["method"],
            significance_level=self.hyperparams["alpha"],
        )

    def _data_loader(self, sample):
        sample_x = sample.astype(np.float32)
        sample_z = np.empty(shape=(sample_x.shape[0], 0)).astype(np.float32)
        dataset = TensorDataset(torch.tensor(sample_x), torch.tensor(sample_z))
        data_loader = DataLoader(
            dataset, batch_size=self.hyperparams["batch_size"], shuffle=False
        )
        return data_loader

    def _train_vae(self, train_loader, valid_loader):
        """Training VAE with the specified image dataset
        :param m: dimension of the latent variable
        :param n: dimension of the observed variable
        :param train_loader: training image dataset loader
        :param valid_loader: validation image dataset loader
        :param biadj_mat: the adjacency matrix of the directed graph
        :param seed: random seed for the experiments
        :return: trained model and training loss history
        """

        m, n = self.doffed.shape

        # building VAE
        mask = self.doffed.T.astype("float32")
        mask = torch.tensor(mask).to(self.device)
        model = VariationalAutoencoder(m, n, mask)
        model = model.to(self.device)
        optimizer = torch.optim.AdamW(
            model.parameters(), lr=self.hyperparams["lr"], weight_decay=1e-5
        )
        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, 50, gamma=0.90)
        num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
        self.log(f"Number of parameters: {num_params}")

        # training loop
        model.train()
        train_elbo, train_error = [], []
        valid_elbo, valid_error = [], []

        for idx in range(self.hyperparams["num_epochs"]):
            self.log(f"Training on epoch {idx}...")
            train_lb, train_er, nbatch = 0.0, 0.0, 0

            for x_batch, _ in train_loader:
                batch_size = x_batch.shape[0]
                x_batch = x_batch.to(self.device)
                recon_batch, logcov_batch, mu_batch, logvar_batch = model(x_batch)
                loss = self._elbo_gaussian(
                    x_batch,
                    recon_batch,
                    logcov_batch,
                    mu_batch,
                    logvar_batch,
                    self.hyperparams["beta"],
                )
                error = self._recon_error(
                    x_batch, recon_batch, logcov_batch, weighted=False
                )
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                # update loss and nbatch
                train_lb += loss.item() / batch_size
                train_er += error.item() / batch_size
                nbatch += 1

            # finish training epoch
            scheduler.step()
            train_lb = train_lb / nbatch
            train_er = train_er / nbatch
            train_elbo.append(train_lb)
            train_error.append(train_er)
            self.log(f"Finish training epoch {idx} with loss {train_lb}")

            # append validation loss
            valid_lb, valid_er = self._valid_vae(model, valid_loader)
            valid_elbo.append(valid_lb)
            valid_error.append(valid_er)

        train_elbo, train_error = np.array(train_elbo), np.array(train_error)
        valid_elbo, valid_error = np.array(valid_elbo), np.array(valid_error)
        elbo = [train_elbo, valid_elbo]
        error = [train_error, valid_error]

        return model, elbo, error

    def _valid_vae(self, model, valid_loader):
        """Training VAE with the specified image dataset
        :param model: trained VAE model
        :param valid_loader: validation image dataset loader
        :return: validation loss
        """
        # set to evaluation mode
        model.eval()
        valid_lb, valid_er, nbatch = 0.0, 0.0, 0

        for x_batch, _ in valid_loader:
            with torch.no_grad():
                batch_size = x_batch.shape[0]
                x_batch = x_batch.to(self.device)
                recon_batch, logcov_batch, mu_batch, logvar_batch = model(x_batch)
                loss = self._elbo_gaussian(
                    x_batch,
                    recon_batch,
                    logcov_batch,
                    mu_batch,
                    logvar_batch,
                    self.hyperparams["beta"],
                )
                error = self._recon_error(
                    x_batch, recon_batch, logcov_batch, weighted=False
                )

                # update loss and nbatch
                valid_lb += loss.item() / batch_size
                valid_er += error.item() / batch_size
                nbatch += 1

        # report validation loss
        valid_lb = valid_lb / nbatch
        valid_er = valid_er / nbatch
        self.log(f"Finish validation with loss {valid_lb}")

        return valid_lb, valid_er

    def _elbo_gaussian(self, x, x_recon, logcov, mu, logvar, beta):
        """Calculating loss for variational autoencoder
        :param x: original image
        :param x_recon: reconstruction in the output layer
        :param logcov: log of covariance matrix of the data distribution
        :param mu: mean in the fitted variational distribution
        :param logvar: log of the variance in the variational distribution
        :param beta: beta
        :return: reconstruction loss + KL
        """

        # KL-divergence
        # https://github.com/AntixK/PyTorch-VAE/blob/master/models/vanilla_vae.py
        # https://github.com/AntixK/PyTorch-VAE/blob/master/models/beta_vae.py
        # https://arxiv.org/pdf/1312.6114.pdf
        kl_div = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())

        # reconstruction loss
        cov = torch.exp(logcov)
        cov = self._apply_along_axis(torch.diag, cov, axis=0)
        cov = cov.mean(axis=0)

        diff = x - x_recon
        recon_loss = torch.sum(
            torch.det(cov)
            + torch.diagonal(
                torch.mm(
                    torch.mm(diff, torch.inverse(cov)), torch.transpose(diff, 0, 1)
                )
            )
        ).mul(-1 / 2)

        # elbo
        loss = -beta * kl_div + recon_loss

        return -loss

    def _recon_error(self, x, x_recon, logcov, weighted):
        """Reconstruction error given x and x_recon
        :param x: original image
        :param x_recon: reconstruction in the output layer
        :param logcov: covariance matrix of the data distribution
        :param weighted: whether to use weighted reconstruction

        Returns
        -------
        error: reconstruction error
        """

        # reconstruction loss
        cov = torch.exp(logcov)
        cov = self._apply_along_axis(torch.diag, cov, axis=0)
        cov = cov.mean(axis=0)

        diff = x - x_recon
        if weighted:
            error = torch.sum(
                torch.det(cov)
                + torch.diagonal(
                    torch.mm(
                        torch.mm(diff, torch.inverse(cov)), torch.transpose(diff, 0, 1)
                    )
                )
            ).mul(-1 / 2)
        else:
            error = torch.linalg.norm(diff, ord=2)

        return error

    @staticmethod
    def _apply_along_axis(function, x, axis=0):
        """Helper function to return along a particular axis
        Parameters
        ----------
        function: function to be applied
        x: data
        axis: axis to apply the function

        Returns
        -------
        The output applied to the axis
        """

        return torch.stack(
            [function(x_i) for x_i in torch.unbind(x, dim=axis)], dim=axis
        )


# implement penalized mle and penalized lse with a new class
class DevMedil(MedilCausalModel):
    def __init__(
        self,
        biadj: npt.NDArray = np.array([]),
        udg: npt.NDArray = np.array([]),
        one_pure_child: bool = True,
        rng=np.random.default_rng(0),
        lambda_reg: float = 0.1,
        mu_reg: float = 0.1,
    ) -> None:
        super().__init__(biadj, udg, one_pure_child, rng)
        self.lambda_reg = lambda_reg
        self.mu_reg = mu_reg

    # penalized MLE
    def fit_penalized_mle(self, dataset: npt.NDArray) -> "DevMedil":
        k, n = self.biadj.shape
        Sigma_hat = np.cov(dataset.T)

        def penalized_mle_loss(params):
            W = params[: k * n].reshape(k, n)
            D = np.diag(params[k * n :])

            Sigma = self.compute_sigma(W, D)
            Sigma_inv = np.linalg.inv(Sigma)
            sign, logdet = np.linalg.slogdet(Sigma_inv)

            if sign <= 0:
                return np.inf

            loss = np.trace(np.dot(Sigma_hat, Sigma_inv)) - logdet
            loss += self.lambda_reg * self.rho(W) + self.mu_reg * self.sigma(W)

            return loss

        initial_W = self.rng.standard_normal((k, n))
        initial_D = self.rng.random(n)
        initial_params = np.hstack([initial_W.flatten(), initial_D])

        result = minimize(penalized_mle_loss, initial_params, method="BFGS")
        self.result = result
        self.W_hat_mle = result.x[: k * n].reshape(k, n)
        self.D_hat_mle = np.diag(result.x[k * n :])
        self.convergence_success_mle = result.success
        self.convergence_message_mle = result.message

        return self

    # penalized LSE
    def fit_penalized_lse(self, dataset: npt.NDArray) -> "DevMedil":
        k, n = self.biadj.shape
        Sigma_hat = np.cov(dataset.T)

        def penalized_lse_loss(params):
            W = params[: k * n].reshape(k, n)
            D = np.diag(params[k * n :])

            loss = norm(Sigma_hat - W.T @ W - D, "fro") ** 2
            # nuclear norm for the first penalty term
            loss += self.lambda_reg * self.rho(W)
            # L1 norm for the second penalty function
            loss += self.mu_reg * self.sigma(W)

            return loss

        initial_W = self.rng.standard_normal((k, n))
        initial_D = np.abs(self.rng.standard_normal(n))
        initial_params = np.concatenate([initial_W.flatten(), initial_D])

        result = minimize(penalized_lse_loss, initial_params, method="BFGS")
        self.result = result
        self.W_hat_lse = result.x[: k * n].reshape(k, n)
        self.D_hat_lse = np.diag(np.abs(result.x[k * n :]))
        self.convergence_success_lse = result.success
        self.convergence_message_lse = result.message

        return self

    # compute sigma
    def compute_sigma(self, W: npt.NDArray, D: npt.NDArray) -> npt.NDArray:
        Sigma = np.dot(W.T, W) + D
        return Sigma

    # ρ(W), the nuclear norm defined as tr(√(W⊤W))
    def rho(self, W: npt.NDArray) -> float:
        WWt = np.dot(W.T, W)  # Compute W^T times W
        sqrt_WWt = sqrtm(WWt)  # Compute the matrix square root of W^T times W
        sqrt_WWt = np.real(sqrt_WWt)  # Take only the real part
        return np.trace(sqrt_WWt)  # Take the trace of the resulting matrix

    # σ(W), the sum of absolute values of elements (L1 norm)
    def sigma(self, W: npt.NDArray) -> float:
        return np.sum(np.abs(W))

    def sample(self, sample_size: int, method: str = "mle") -> npt.NDArray:
        if method not in ["mle", "lse"]:
            raise ValueError("Method must be either 'mle' or 'lse'")

        if method == "mle":
            if not hasattr(self, "W_hat_mle") or not hasattr(self, "D_hat_mle"):
                raise ValueError("MLE model must be fitted before sampling")
            W_hat, D_hat = self.W_hat_mle, self.D_hat_mle
        else:
            if not hasattr(self, "W_hat_lse") or not hasattr(self, "D_hat_lse"):
                raise ValueError("LSE model must be fitted before sampling")
            W_hat, D_hat = self.W_hat_lse, self.D_hat_lse

        k, n = W_hat.shape
        L = self.rng.standard_normal((sample_size, k))
        epsilon = self.rng.multivariate_normal(np.zeros(n), D_hat, sample_size)
        return np.dot(L, W_hat) + epsilon

    def fit(self, dataset: npt.NDArray, method: str = "mle") -> "DevMedil":
        if method == "mle":
            return self.fit_penalized_mle(dataset)
        elif method == "lse":
            return self.fit_penalized_lse(dataset)
        else:
            raise ValueError("Method must be either 'mle' or 'lse'")



"""Generate random minimum MeDIL causal model graph or parameters."""
import numpy as np
import numpy.typing as npt
from numpy.random import default_rng

from .models import GaussianMCM
from .ecc_algorithms import find_clique_min_cover


def mcm(
    rng: np.random.Generator = default_rng(0),
    parameterization: str = "Gaussian",
    biadj: npt.NDArray = np.array([]),
    **kwargs,
) -> GaussianMCM:
    if biadj.size == 0:
        biadj = _biadj(**kwargs)
    if parameterization == "Gaussian":
        mcm = GaussianMCM(biadj=biadj)
        params = mcm.parameters

        num_edges = biadj.sum()
        weights = (rng.random(num_edges) * 1.5) + 0.5
        weights[rng.choice((True, False), num_edges)] *= -1
        params.biadj_weights = np.zeros_like(biadj, float)
        params.biadj_weights[biadj] = weights

        num_meas = biadj.shape[1]
        params.error_means = rng.random(num_meas) * 2
        params.error_means[rng.choice((True, False), num_meas)] *= -1

        params.error_variances = (rng.random(num_meas) * 1.5) + 0.5

    elif parameterization == "VAE":
        raise (NotImplementedError)

    else:
        raise ValueError(f"Parameterization '{parameterization}' is invalid.")

    return mcm


def biadj(
    num_meas: int,
    density: float = 0.2,
    one_pure_child: bool = True,
    num_latent: int = 0,
    rng: np.random.Generator = default_rng(0),
) -> npt.NDArray:
    """Randomly generate biadjacency matrix for graphical minMCM."""
    if one_pure_child:
        """Define a maximum independent set of size `num_latent`, and then grow these into a minimum edge clique cover with average max clique size `2 + (num_meas - num_latent) * density`."""
        if num_latent == 0:
            num_latent = rng.integers(1, num_meas)
        if density is None:
            density = rng.random()

        # specify pure children/independent set
        biadj = np.zeros((num_latent, num_meas), bool)
        biadj[:, :num_latent] = np.eye(num_latent)

        # every child gets a parent; specifically L_0, until the
        # within-column perm below using np.permuted
        biadj[0, num_latent:] = True

        # randomly fill in remaining density * (num_meas - num_latent)
        # * (num_latent - 1) edges
        max_num_edges = (num_meas - num_latent) * (num_latent - 1)
        num_edges = np.round(max_num_edges * density).astype(int)

        edges = np.zeros(max_num_edges, bool)
        edges[:num_edges] = True
        edges = rng.permutation(edges).reshape(num_latent - 1, num_meas - num_latent)

        biadj[1:][:, num_latent:] = edges

        nonpure_children = biadj[:, num_latent:]
        biadj[:, num_latent:] = rng.permuted(nonpure_children, axis=0)

        # change child order, so pure children aren't first
        biadj = rng.permutation(biadj, axis=1)

    else:
        """Generate minMCM from Erdős–Rényi random undirected graph
        over observed variables."""
        if num_latent != 0:
            msg = "`num_latent` can only be specified when `one_pure_child==True`."
            raise ValueError(msg)

        udg = np.zeros((num_meas, num_meas), bool)

        max_edges = (num_meas * (num_meas - 1)) // 2
        num_edges = np.round(density * max_edges).astype(int)

        edges = np.ones(max_edges)
        edges[num_edges:] = 0

        udg[np.triu_indices(num_meas, k=1)] = rng.permutation(edges)
        udg += udg.T
        np.fill_diagonal(udg, True)

        # find latent connections (minimum edge clique cover)
        biadj = find_clique_min_cover(udg).astype(bool)

    return biadj


_biadj = biadj



from setuptools import setup


def long_description():
    with open("README.md") as readme:
        return readme.read()


setup(
    name="medil",
    version="1.1.0",
    author="Alex Markham",
    author_email="alex.markham@causal.dev",
    description="MeDIL is a Python package for causal factor analysis with the (me)asurement (d)ependence (i)nducing (l)atent causal model framework.",
    long_description_content_type="text/markdown",
    long_description=long_description(),
    license="GNU Affero General Public License version 3 or later (AGPLv3+)",
    packages=["medil"],
    install_requires=[
        "dcor",
        "matplotlib",
        "networkx",
        "seaborn",
        "scikit-learn",
        "torch",
    ],
)



import itertools

import numpy as np
import numpy.typing as npt


def sfd(predicted_biadj, true_biadj):
    """Perform analysis of the distances between true and reconstructed structures
    Parameters
    ----------
    biadj_mat: input directed graph
    biadj_mat_recon: learned directed graph in the form of adjacency matrix

    Returns
    -------
    sfd: squared Frobenius distance (bipartite graph)
    ushd: structural hamming distance (undirected graph)
    """

    # ushd = shd_func(recover_ug(biadj_mat), recover_ug(biadj_mat_recon))
    ug = recover_ug(true_biadj)
    ug_recon = recover_ug(predicted_biadj)

    ushd = np.triu(np.logical_xor(ug, ug_recon), 1).sum()

    true_biadj = true_biadj.astype(int)
    predicted_biadj = predicted_biadj.astype(int)

    wtd_ug = true_biadj.T @ true_biadj
    wtd_ug_recon = predicted_biadj.T @ predicted_biadj

    sfd = ((wtd_ug - wtd_ug_recon) ** 2).sum()

    return sfd, ushd


def recover_ug(biadj_mat):
    """Recover the undirected graph from the directed graph
    Parameters
    ----------
    biadj_mat: learned directed graph

    Returns
    -------
    ug: the recovered undirected graph
    """

    # get the undirected graph from the directed graph
    ug = biadj_mat.T @ biadj_mat
    np.fill_diagonal(ug, False)

    return ug


def min_perm_squared_l2_dist(predicted_W: npt.NDArray, true_W: npt.NDArray):
    def perm_squared_l2_dist(perm):
        return np.sum((predicted_W[perm] - true_W) ** 2)

    def pair(perm):
        return perm, perm_squared_l2_dist(perm)

    perms = itertools.permutations(range(len(predicted_W)))

    pairs = map(pair, perms)

    opt_perm, min_dist = min(pairs, key=lambda pair: pair[1])

    return opt_perm, min_dist



from itertools import permutations

import pytest
import numpy as np

from medil.models import MedilCausalModel, GaussianMCM, NeuroCausalFactorAnalysis


class TestMedilCausalModel:
    def test_base(self):
        mcm = MedilCausalModel()
        with pytest.raises(NotImplementedError):
            mcm.fit(np.array([]))
        with pytest.raises(NotImplementedError):
            mcm.sample(0)


class TestGaussianMCM:
    def test_sample_m(self):
        """Simple "M" graph, with 2 latent and 3 measurement vars."""
        biadj = np.zeros((2, 3), bool)
        biadj[[0, 0, 1, 1], [0, 1, 1, 2]] = True
        mcm = GaussianMCM(biadj=biadj)
        params = mcm.parameters
        params.biadj_weights = biadj.astype(float)
        params.error_means = np.zeros(3)
        params.error_variances = np.ones(3)

        s = mcm.sample(10000)
        assert np.allclose(s.mean(0), mcm.parameters.error_means, atol=0.02)

    def test_sample_empty(self):
        """When UDG is empty graph."""
        biadj = np.eye(5, dtype=bool)
        mcm = GaussianMCM(biadj=biadj)
        params = mcm.parameters
        params.biadj_weights = biadj.astype(float)
        params.error_means = np.zeros(5)
        params.error_variances = np.ones(5)

        dataset = mcm.sample(100000)
        assert np.allclose(dataset.mean(0), mcm.parameters.error_means, atol=0.02)

    def test_fit_m(self):
        """Simple "M" graph, with 2 latent and 3 measurement vars."""
        biadj = np.zeros((2, 3), bool)
        biadj[[0, 0, 1, 1], [0, 1, 1, 2]] = True
        mcm = GaussianMCM(biadj=biadj)
        params = mcm.parameters
        params.biadj_weights = biadj.astype(float)
        params.error_means = np.zeros(3)
        params.error_variances = np.ones(3)

        dataset = mcm.sample(10000)

        mcm_est = GaussianMCM().fit(dataset)
        params_est = mcm_est.parameters
        assert (mcm.biadj == mcm_est.biadj).all()
        assert np.allclose(params_est.biadj_weights, params.biadj_weights, atol=0.02)
        assert np.allclose(params_est.error_means, params.error_means, atol=0.02)
        assert np.allclose(
            params_est.error_variances, params.error_variances, atol=0.02
        )

    def test_fit_random(self):
        """Randomly generated MCM."""
        biadj = np.array(
            [
                [False, False, False, True, False],
                [True, True, False, False, False],
                [False, False, True, False, False],
                [True, False, False, False, True],
            ]
        )
        mcm = GaussianMCM(biadj=biadj)
        params = mcm.parameters
        params.biadj_weights = np.array(
            [
                [0.0, 0.0, 0.0, 1.45544253, 0.0],
                [0.90468007, 0.56146029, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.52479145, 0.0, 0.0],
                [1.71990536, 0.0, 0.0, 0.0, 1.86913337],
            ]
        )
        params.error_means = np.array(
            [1.87014485, 1.63170711, 0.005477, -1.71480855, -0.06717115]
        )
        params.error_variances = np.array(
            [1.31219183, 0.94956784, 1.13403083, 0.54247951, 0.68642491]
        )

        dataset = mcm.sample(10000)

        mcm_est = GaussianMCM().fit(dataset)
        params_est = mcm_est.parameters

        udg = biadj.T @ biadj
        np.fill_diagonal(udg, False)
        assert (udg == mcm_est.udg).all()

        for p in permutations(range(4)):
            p = np.array(p)
            if (mcm_est.biadj[p] == biadj).all():
                break
        assert (mcm_est.biadj[p] == biadj).all()

        assert np.allclose(params_est.biadj_weights[p], params.biadj_weights, atol=0.5)

        assert np.allclose(params_est.error_means, params.error_means, atol=0.05)
        assert np.allclose(params_est.error_variances, params.error_variances, atol=0.7)

    # def test_ncfa_assign_dof(self):
    #     biadj_mat = np.array([[0, 0, 1, 1, 1], [0, 1, 0, 1, 0], [1, 0, 1, 0, 0]])
    #     variances = np.array([2.5, 0.33, 2.5, 0.66, 0.88])

    #     warnings.filterwarnings("error")
    #     try:
    #         test_insufficient = assign_DoF(biadj_mat, 2, "uniform")
    #         assert False
    #     except UserWarning:
    #         warnings.resetwarnings()
    #         warnings.simplefilter("ignore")
    #         test_insufficient = assign_DoF(biadj_mat, 2, "uniform")
    #         assert (test_insufficient == biadj_mat).all()

    #     test_uniform = assign_DoF(biadj_mat, 8, "uniform")
    #     unique_uniform, counts_uniform = np.unique(
    #         test_uniform, axis=0, return_counts=True
    #     )
    #     assert (biadj_mat == unique_uniform).all()
    #     assert min(counts_uniform) == 2
    #     assert max(counts_uniform) == 3
    #     assert counts_uniform.sum() == 8

    #     test_clique = assign_DoF(biadj_mat, 11, "clique_size")
    #     unique_clique, counts_clique = np.unique(
    #         test_clique, axis=0, return_counts=True
    #     )
    #     assert (biadj_mat == unique_clique).all()
    #     assert min(counts_clique) == 3
    #     assert max(counts_clique) == 4
    #     assert counts_clique.sum() == 11

    #     test_tot = assign_DoF(biadj_mat, 13, "tot_var", variances)
    #     unique_tot, counts_tot = np.unique(test_tot, axis=0, return_counts=True)
    #     assert (biadj_mat == unique_tot).all()
    #     assert ((5, 2, 6) == counts_tot).all()

    #     test_avg = assign_DoF(biadj_mat, 29, "avg_var", variances)
    #     unique_avg, counts_avg = np.unique(test_avg, axis=0, return_counts=True)
    #     assert (biadj_mat == unique_avg).all()
    #     assert ((9, 4, 16) == counts_avg).all()

    #     for dof in range(3, 12):
    #         for method in ("uniform", "clique_size", "tot_var", "avg_var"):
    #             test_rounding = assign_DoF(biadj_mat, dof, method, variances)
    #             assert (np.unique(test_rounding, axis=0) == biadj_mat).all()
    #             assert dof == len(test_rounding)


class TestNeuroCausalFactorAnalysis:
    def test_fit_m_gaussian(self):
        """Simple "M" graph, with 2 latent and 3 measurement vars, sampled from GaussianMCM."""
        biadj = np.zeros((2, 3), bool)
        biadj[[0, 0, 1, 1], [0, 1, 1, 2]] = True
        mcm = GaussianMCM(biadj=biadj)
        params = mcm.parameters
        params.biadj_weights = biadj.astype(float)
        params.error_means = np.zeros(3)
        params.error_variances = np.ones(3)

        dataset = mcm.sample(10000)

        NeuroCausalFactorAnalysis(verbose=True).fit(dataset)



"""Independence testing on samples of random variables."""

from typing import NamedTuple, Optional

from dcor.independence import distance_correlation_t_test
from multiprocessing import Pool, cpu_count
import numpy as np
import numpy.typing as npt
from scipy.spatial.distance import pdist, squareform
from scipy.stats import chi2, norm, rankdata


def dcov(samples):
    r"""Compute sample distance covariance matrix.
    Parameters
    ----------
    samples : 2d numpy array of floats
              A :math:`N \times M` matrix with :math:`N` samples of
              :math:`M` random variables.
    Returns
    -------
    2d numpy array
        A square matrix :math:`C`, where :math:`C_{i,j}` is the sample
        distance covariance between random variables :math:`R_i` and
        :math:`R_j`.
    """
    num_samps, num_feats = samples.shape
    num_pairs = num_samps * (num_samps - 1) // 2
    dists = np.zeros((num_feats, num_pairs))
    d_bars = np.zeros(num_feats)
    # compute doubly centered distance matrix for every feature:
    for feat_idx in range(num_feats):
        n = num_samps
        t = np.tile
        # raw distance matrix:
        d = squareform(pdist(samples[:, feat_idx].reshape(-1, 1), "cityblock"))
        # doubly centered:
        d_bar = d.mean()
        d -= t(d.mean(0), (n, 1)) + t(d.mean(1), (n, 1)).T - t(d_bar, (n, n))
        d = squareform(d, checks=False)  # ignore assymmetry due to numerical error
        dists[feat_idx] = d
        d_bars[feat_idx] = d_bar
    return dists @ dists.T / num_samps**2, d_bars


def estimate_UDG(sample, method="dcov_fast", significance_level=0.05):
    samp_size, num_feats = sample.shape

    if isinstance(method, np.ndarray):
        p_vals = method
        udg = p_vals < significance_level
    elif method == "dcov_fast":
        cov, d_bars = dcov(sample)
        crit_val = chi2(1).ppf(1 - significance_level)
        test_val = samp_size * cov / np.outer(d_bars, d_bars)
        udg = test_val >= crit_val
        p_vals = None
    elif method == "g-test":
        pass
    else:
        p_vals = np.zeros((num_feats, num_feats), float)
        idxs, jdxs = np.triu_indices(num_feats, 1)
        zipped = zip(idxs, jdxs)
        sample_iter = (sample[:, i_j].T for i_j in zipped)
        if method == "dcov_big":
            test = dcor_test
        elif method == "xicor":
            test = xicor_test
        with Pool(max(1, int(0.75 * cpu_count()))) as p:
            p_vals[idxs, jdxs] = p_vals[jdxs, idxs] = np.fromiter(
                p.imap(test, sample_iter, 100), float
            )
            udg = p_vals < significance_level
    np.fill_diagonal(udg, False)
    return udg, p_vals


def dcor_test(x_y):
    x, y = x_y
    return distance_correlation_t_test(x, y).pvalue


def xicor_test(x_y):
    x, y = x_y
    xi, pvalue = xicorr(x, y)
    return pvalue


# The following is a modification of the xicorrelation source code,
# Copyright 2021 Nikolay Novik (https://github.com/jettify),
# for compatibility with numpy 2.0+
class _XiCorr(NamedTuple):
    xi: float
    fr: npt.NDArray[np.float64]
    cu: float


class XiCorrResult(NamedTuple):
    correlation: float
    pvalue: Optional[float]


def _xicorr(x: npt.NDArray[np.float64], y: npt.NDArray[np.float64]) -> _XiCorr:
    # Ported from original R implementation.
    # https://github.com/cran/XICOR/blob/master/R/calculateXI.R
    n = x.size
    PI = rankdata(x, method="average")
    fr = rankdata(y, method="average") / n
    gr = rankdata(-y, method="average") / n
    cu = np.mean(gr * (1 - gr))
    A1 = np.abs(np.diff(fr[np.argsort(PI, kind="quicksort")])).sum() / (2 * n)
    xi = 1.0 - A1 / cu
    return _XiCorr(xi, fr, cu)


def xicorr(x: npt.ArrayLike, y: npt.ArrayLike, ties: bool = True) -> XiCorrResult:
    """Compute the cross rank increment correlation coefficient xi [1].

    Parameters
    ----------
    x, y : array_like
        Arrays of rankings, of the same shape. If arrays are not 1-D, they
        will be flattened to 1-D.
    ties : bool, optional
        If ties is True, the algorithm assumes that the data has ties and
        employs the more elaborated theory for calculating s.d. and P-value.
        Otherwise, it uses the simpler theory. There is no harm in putting
        ties = True even if there are no ties.

    Returns
    -------
    correlation : float
       The tau statistic.
    pvalue : float
       P-values computed by the asymptotic theory.

    See Also
    --------
    spearmanr : Calculates a Spearman rank-order correlation coefficient.


    Example:
        >>> from xicorrelation import xicorr
        >>> x = [1, 2, 3, 4, 5]
        >>> y = [1, 4, 9, 16, 25]
        >>> xi, pvalue = xicorr(x, y)
        >>> print(xi, pvalue)

    References
    ----------
    .. [1] Chatterjee, S., "A new coefficient of correlation",
           https://arxiv.org/abs/1909.10140, 2020.

    Examples
    --------
    >>> from scipy import stats
    >>> x1 = [12, 2, 1, 12, 2]
    >>> x2 = [1, 4, 7, 1, 0]
    >>> xi, p_value, _ = xicorr(x1, x2)
    >>> tau
    -0.47140452079103173
    >>> p_value
    0.2827454599327748
    """
    # https://git.io/JSIlN
    x = np.asarray(x).ravel()
    y = np.asarray(y).ravel()

    if x.size != y.size:
        raise ValueError(
            "All inputs to `xicorr` must be of the same "
            f"size, found x-size {x.size} and y-size {y.size}"
        )
    elif not x.size or not y.size:
        # Return NaN if arrays are empty
        return XiCorrResult(np.nan, np.nan)

    r = _xicorr(x, y)
    xi = r.xi
    fr = r.fr
    CU = r.cu

    pvalue: Optional[float] = None
    # https://git.io/JSIlM
    n = x.size
    if not ties:
        # sd = np.sqrt(2.0 / (5.0 * n))
        pvalue = 1.0 - norm.cdf(np.sqrt(n) * xi / np.sqrt(2.0 / 5.0))
    else:
        qfr = np.sort(fr)
        ind = np.arange(1, n + 1)
        ind2 = 2 * n - 2 * ind + 1

        ai = np.mean(ind2 * qfr * qfr) / n
        ci = np.mean(ind2 * qfr) / n
        cq = np.cumsum(qfr)
        m = (cq + (n - ind) * qfr) / n
        b = np.mean(m**2)
        v = (ai - 2.0 * b + ci**2) / (CU**2)

        # sd = np.sqrt(v / n)
        pvalue = 1.0 - norm.cdf(np.sqrt(n) * xi / np.sqrt(v))
    return XiCorrResult(xi, pvalue)
